#ifndef SYSTEM_H
#define SYSTEM_H

#include <fstream>
#include <complex>
#include <stdlib.h>
#include "tensor.h"
#include "global.h"
#include "MLSLv3.h"
#include "McLatticePr.h"

#define MAX_NUM_BLOCK 100;

using namespace std;

const val_type PI = 3.1415926535897932384626433832795;
typedef complex<val_type> cplx;

void fft(cplx *x, int sz);

void fft_error(cplx *x, int sz);

void calc_psd(val_type *in, val_type *out, int size);

class Block {
public:
	static int number_of_blocks;
	Block() { 
		  name = NULL; number_of_blocks++; 
		  sprintf(aux_string, "Number of blocks %d.\n", number_of_blocks);
		  ANCDebugOutput(aux_string);
	};
      virtual ~Block() {
		  number_of_blocks--;
		  sprintf(aux_string, "%d blocks left.\n", number_of_blocks);
		  ANCDebugOutput(aux_string);
	  };

	char *name;
	virtual void calc_output() {};
	virtual void clock_tick() {};
	virtual void save() {};
    virtual char *getName() { return name; }
};

ostream & operator <<(ostream &out, Tensor *t);

typedef Block *BlockPointer;

class Buffer: public Block {
public:
	Tensor *buf;
	Tensor *input;
	int size;

	// The buffer depth is defined by the first index

	Buffer(int s, Tensor *in, char *n = NULL):size(s), input(in) {
		int dim = in->dim+1;
		int *size_vector= new int[dim];
		size_vector[0]=size;
		for(int i=1; i<dim; i++) size_vector[i]=in->size_vector[i-1];
		buf = new Tensor(size_vector, dim);
		memset(buf->values, 0, buf->size*sizeof(val_type));
		name = n;
		delete size_vector;
	};

	virtual ~Buffer() {
            ANCDebugOutput("Destroying buffer\n");
		delete buf;
	}

	void clock_tick();
	void calc_output();

	void save() {
		if (name==NULL) return;

		ofstream stream;
		char filename[100];
		strcpy(filename, "Output/");
		strcat(filename, name);
		strcat(filename, "_Buffer.txt");

		stream.open(filename);
		stream<<buf;
		stream.close();
	}
};

class WhiteNoiseGenerator :public Block {
public:
	val_type amplitude;
	Tensor *output;
	WhiteNoiseGenerator(val_type amp, int size) {
		output = new Tensor(size);
		amplitude = amp;
	}
	void calc_output() {
		for(int i=0; i<output->getSize(); i++)
			output->values[i] = amplitude*(((double)rand())/RAND_MAX-0.5);
	}

	virtual ~WhiteNoiseGenerator() {
		ANCDebugOutput("Destroying WhiteNoiseGenerator\n");
		delete output;
	}
};

class MFIR;

class FIR: public Block {
public:
	Tensor *coeficients;
	Buffer *buffer;
	Tensor *output;
	bool	link;

	FIR(int s, int in_s, int out_s, Tensor *in, char *n=NULL) {
		ASSERT(in->getDim()==1 && in->getSize()==in_s);
		buffer = new Buffer(s, in, n);
		coeficients = new Tensor(s, in_s, out_s);
		output = new Tensor(out_s);
		link = false;
		name = n;
	}

	FIR(FIR *f, Tensor *in) {
		ASSERT(in->getDim()==1 && f->coeficients->getDim()==3 &&
			f->coeficients->getSize(1)==in->getSize(0));
		buffer = new Buffer(f->coeficients->getSize(0), in);
		coeficients = f->coeficients;
		output = new Tensor(f->output->getSize(0));
		link = true;
	}

	FIR() {
		buffer = NULL;
		coeficients = NULL;
		output = NULL;
		link = false;
	}

	virtual ~FIR(){
        ANCDebugOutput("Destroying FIR\n");
		if (!link) delete coeficients;
		delete buffer;
		delete output;
	}

	void calc_output() {
		buffer->calc_output();
		DotT(buffer->buf, coeficients, 2, output);
	}

	void clock_tick() { buffer->clock_tick(); }

	void SubSampledFIR(FIR *fir, MFIR *anti_aliasing, MFIR *reconstruction, 
		int sampling_time, Tensor *in);

	void save() {
		if (name==NULL) return;

		ofstream stream;
		char filename[100];
		strcpy(filename, "Output/");
		strcat(filename, name);
		strcat(filename, "_FIR.txt");

		stream.open(filename);
		stream<<coeficients;
		stream.close();
	}
};

class MFIR: public Block {
public:
	Tensor *coeficients;
	Buffer *buffer;
	Tensor *output;
	bool link;

	MFIR() { coeficients=NULL; buffer=NULL; output=NULL; };
	MFIR(int s, Tensor *in) {
		ASSERT(in->getDim()==1);
		buffer = new Buffer(s, in);
		coeficients = new Tensor(s);
		for(int i=0; i<coeficients->getSize(); i++)
			coeficients->values[i]=0;
		output = new Tensor(in->getSize());
		link = false;
	}
	MFIR(MFIR *f, Tensor *in) {
		buffer = new Buffer(f->coeficients->getSize(0), in);
		coeficients = f->coeficients;
		output = new Tensor(in->getSize());
		link = true;
	}

	void SubSampledFIR(MFIR *fir, MFIR *anti_aliasing, MFIR *reconstruction, 
						 int sampling_time, Tensor *in);

	void LowPass(double wc);

	virtual ~MFIR(){
		ANCDebugOutput("Destroying MFIR\n");
		if (!link) delete coeficients;
		delete buffer;
		delete output;
	}

	void calc_output() {
        buffer->calc_output();
		for(int j=0; j<output->getSize(); j++) {
			Matrix_MultiplyT(buffer->buf->values + j*(buffer->size), 
				coeficients->values, output->values+j, 
				buffer->size , 1, 1);
		}
	}

	void clock_tick() {
		buffer->clock_tick();
	}
};

class Delay: public MFIR {
public:
	Delay(int delay, Tensor *in):MFIR(delay+1, in) {
		coeficients->pos(delay)=1;
	}
	virtual ~Delay() {
        ANCDebugOutput("Destroying Delay\n");
	}
};

class Sum :public Block {
public:
	typedef  Tensor *TensorPtr;
	TensorPtr inputs[10];
	Tensor *output;
	int n_inputs;

	Sum(Tensor *in1, Tensor *in2){
		ASSERT(in1->dim==1 && in2->dim==1 && in1->getSize(0)==in2->getSize(0));
		n_inputs = 2;
		output = new Tensor(in1->size_vector, in1->dim);
		inputs[0] = in1;
		inputs[1] = in2;
	}

	void add_input(Tensor *in) {
		ASSERT(n_inputs<10-1);
		ASSERT(in->dim==1 && in->getSize(1)==inputs[0]->getSize(0));
		inputs[n_inputs++]=in;
	}
	virtual ~Sum() {
        ANCDebugOutput("Destroying Sum\n");
		delete output;
	}

	void calc_output() {
		Plus(inputs[0], inputs[1], output);
		for(int i=2; i<n_inputs; i++)
			Plus(inputs[i], output, output);
	}
};

class Difference :public Block {
public:
	Tensor *input1;
	Tensor *input2;
	Tensor *output;

	Difference(Tensor *in1, Tensor *in2) {
		ASSERT(in1->dim==1 && in2->dim==1 && in1->getSize(0)==in2->getSize(0));
		input1 = in1;
		input2 = in2;
		output = new Tensor(in1->getSize(0));
	}

	void calc_output() {
		Minus(input1, input2, output);
	}

	virtual ~Difference() {
        ANCDebugOutput("Destroying Difference\n");
		delete output;
	}
};

class Gain: public Block {
public:
	Tensor *input;
	Tensor *output;
	val_type value;

	Gain(Tensor *in1, val_type val) {
		input = in1;
		output = new Tensor(in1->getSize(0));
		value = val;
	}

	void calc_output() {
		Times(value, input, output);
	}

	virtual ~Gain() {
        ANCDebugOutput("Destroying Gain\n");
		delete output;
	}
};

int get_n_channels(Tensor *x);

class PSD: public Block {
public:
	Tensor *input;
	Tensor *output;
	Tensor *aux;
	int n_channels;
	int time;

	PSD(Tensor *in) {
		output = new Tensor(in->size_vector, in->dim);
		input = in;
		n_channels = get_n_channels(in);
		aux = new Tensor(in->size_vector, in->dim);
		time = 0;
	}

	virtual ~PSD() {
		ANCDebugOutput("Destroying PSD\n");
		delete output;
		delete aux;
	}

	void clock_tick(){

		if (time++ >= input->getSize(0)/10) {
			time = 0;
			for(int i=0; i<n_channels; i++) {
				calc_psd(input->values+i*input->getSize(0), 
					aux->values+i*input->getSize(0), 
					input->getSize(0));
			}

			Times(1.0/100, aux, aux);
			Times(1.0-1.0/100, output, output);
			Plus(aux, output, output);
		}
	}

	void save() {
		if (name==NULL) return;

		ofstream stream;
		char filename[100];
		strcpy(filename, "Output/");
		strcat(filename, name);
		strcat(filename, "_PSD.txt");

		stream.open(filename);
		stream<<output;
		stream.close();
	}

};

class BlockSet :public Block {
public:
	virtual void add(BlockPointer b) {};
	virtual void add2(BlockPointer b) {};
    virtual void list() {};
};

typedef Block *BlockPointer;

class SubSystem :public BlockSet {
public:
	SubSystem();
	virtual ~SubSystem();

	void calc_output();
	void clock_tick();

	BlockPointer *blocks;
	bool *links_to_block;

	int num_blocks;
	int max_num_blocks;

	void add(BlockPointer b);
	void add2(BlockPointer b);
	void save();
    virtual void list();
};

class SubSampledSystem :public SubSystem {
public:
	SubSampledSystem(int sample_time);
	int sample_time;
	int current_time;
	bool enable;

	void calc_output();
	void clock_tick();
};

class ImpulseSampler: public Block {
public:
	Tensor *input;
	Tensor *output;
	int sampling_time;
	int clock;
	ImpulseSampler(int time, Tensor *in){
		input = in;
		sampling_time = time;
		output = new Tensor(in->size_vector, in->dim);
		clock = 0;
	}

	virtual ~ImpulseSampler() {
        ANCDebugOutput("Destroying ImpulseSampler\n");
		delete output;
	}

	void clock_tick() {
		++clock;
		if (clock >= sampling_time) clock = 0;
	}
	void calc_output() {
		if (clock==0) Times(sampling_time, *input, *output);
		else Times(0, *input, *output);
	}
};

class RFIR: public MFIR {
public:
	ImpulseSampler *imp_spl;

	RFIR(int s, Tensor *in, int time): 
		MFIR(s, in) {
		imp_spl = new ImpulseSampler(time, in);
		buffer->input = imp_spl->output;
	}

	RFIR(MFIR *f, Tensor *in, int time):
		MFIR(f, in) {
		imp_spl = new ImpulseSampler(time, in);
		buffer->input = imp_spl->output;
	}

	virtual ~RFIR(){
        ANCDebugOutput("Destroying RFIR\n");
		delete imp_spl;
	}

	void calc_output() {
		imp_spl->calc_output();
		MFIR::calc_output();
	}

	void clock_tick() {
		imp_spl->clock_tick();
		MFIR::clock_tick();
	}
};

class Power: public Block {
public:
	Tensor *input;
	Tensor *output;
	Tensor *aux;
	val_type alpha;

	Power(val_type time, Tensor *in) {
		input = in;
		output = new Tensor(input->size_vector, input->dim);
		aux = new Tensor(input->size_vector, input->dim);
		alpha = 1/time;
	}

	virtual ~Power() {
        ANCDebugOutput("Destroying Power\n");
		delete output;
 		delete aux;
	}

	void clock_tick() {
		// output = (1-alpha)*output + alpha*input*input;
		Element_Multiply(input->values
			, input->values, 
			aux->values, 
			input->size);

		Times(alpha, aux, aux);
		Times(1-alpha, output, output);
		Plus(aux, output, output);
	}
};

class LMS: public Block {
public:
	FIR *fir;

	Tensor *desired_signal;
	Tensor *error_signal;
	Tensor *aux;
	val_type alpha;

	LMS(FIR *f, Tensor *des_sig) {
			fir = f;
			desired_signal=des_sig;
			aux = new Tensor(fir->coeficients->size_vector, fir->coeficients->dim);
			error_signal = new Tensor(desired_signal->size_vector, desired_signal->dim);
			alpha = 0.5;
	}

	virtual ~LMS() {
        ANCDebugOutput("Destroying LMS\n");
		delete aux;
		delete error_signal;
	}
	void calc_output(){
		val_type mu;

		Minus(desired_signal, fir->output, error_signal);
		mu = alpha/(calc_power(fir->buffer->buf)+1e-9);
		Dot(fir->buffer->buf, error_signal, 0, aux);
		Times(mu, aux, aux);
		Plus(fir->coeficients, aux, fir->coeficients);
	}
};

class LSLBlk: public Block {
public:
	MLSLattice *lattice;

	Tensor *desired_signal;
	Tensor *input;
	Tensor *output;
	Tensor *refection_coefficients;

	LSLBlk(int s, int in_s, int out_s, Tensor *in, Tensor *des, val_type val=0, char *n=NULL) {
		lattice = new MLSLattice(in_s, out_s, s-1, 0.999, 0.01);
		// time constant ~= 1/(1-lambda)=1000
		ASSERT(in_s==in->getSize(0));
		ASSERT(in->getDim()==1);
		ASSERT(des->getSize(0)==out_s);
		ASSERT(des->getDim()==1);
		input = in;
		desired_signal=des;
		output = new Tensor(out_s);
		int aux[3]= {out_s, in_s, s};
		refection_coefficients = 
			new Tensor(aux, 3, lattice->kappa0p);
	}

	~LSLBlk() {
		ANCDebugOutput("Destroying LSLBlk\n");
		delete lattice;
		delete output;
		delete refection_coefficients;
	}

	void calc_output(){
		lattice->Update(input->values, desired_signal->values);
		lattice->Output(output->values);
	}

	void clock_tick(){
	}
};

class LatticeBlk: public Block {
public:
	McLatticePr *lattice;

	Tensor *input;
	Tensor *output;

	LatticeBlk(Tensor *in, int out_s, char *n=NULL) {

		input = in;
		output = new Tensor(out_s);
		lattice = NULL;
	}

	void Link(LSLBlk *link){
		lattice = new McLatticePr(link->lattice->inSize, link->lattice->outSize, link->lattice->order,  
		link->lattice->GammaF0p, link->lattice->GammaB0p, link->lattice->kappa0p, link->lattice->gamma0p);
	}

	~LatticeBlk() {
		ANCDebugOutput("Destroying LatticeBlk\n");
		delete lattice;
		delete output;
	}

	void calc_output(){
		lattice->Update(input->values);
		lattice->Output(output->values);
	}

	void clock_tick(){
	}
};




class FXLMS: public Block {
public:
	Tensor *reference;
	Buffer *reference_buffer;
	Tensor *filtered_reference;
	Buffer *filtered_reference_buffer;
	Tensor *desired_signal;
      Tensor *coeficients;
      Tensor *cancelation_path;

	Tensor *aux;
	Tensor *error_signal;
	Tensor *aux2;

	val_type alpha;
	val_type lambda;

	FXLMS(Tensor *ref, Tensor *des_sig, Tensor *coef, Tensor *canc_path) {
			reference = ref;
			desired_signal = des_sig;
			coeficients = coef;
			reference_buffer = 
				new Buffer(canc_path->getSize(0), reference, "Reference_buffer");

			filtered_reference = new Tensor(ref->getSize(0), 
				canc_path->getSize(1),
				canc_path->getSize(2)
				);
			filtered_reference_buffer = 
				new Buffer(coef->getSize(0), filtered_reference);

			cancelation_path = canc_path;

			aux = new Tensor(coeficients->size_vector, coeficients->dim);
			error_signal = new Tensor(desired_signal->size_vector, desired_signal->dim);
			aux2 = new Tensor(error_signal->size_vector, error_signal->dim);

			alpha = 0.3;
			lambda = 0.9999;
	}

	virtual ~FXLMS() {
        ANCDebugOutput("Destroying FXLMS\n");
		delete reference_buffer;
		delete filtered_reference;
		delete filtered_reference_buffer;
		delete aux;
		delete error_signal;
		delete aux2;
	}

	void calc_output(){
		val_type mu;

		reference_buffer->calc_output();
		DotT(reference_buffer->buf, cancelation_path, 1, filtered_reference);
		filtered_reference_buffer->calc_output();

 		DotT(filtered_reference_buffer->buf, coeficients, 3, aux2);
		Minus(desired_signal, aux2, error_signal);

		mu = alpha/(calc_power(filtered_reference_buffer->buf)+1e-3);
		Dot(filtered_reference_buffer->buf, error_signal, 1, aux);
		Times(mu, aux, aux);
		Times(lambda, coeficients, coeficients);
		Plus(coeficients, aux, coeficients);
	}

	void clock_tick(){
		filtered_reference_buffer->clock_tick();
		reference_buffer->clock_tick();
	}
};

class IOBlock: public Block {
public:
	Tensor *reference; 	// The references from the PCM16IO
	Tensor *error; 		// The error signals from the PCM16IO
	Tensor *anti_noise;  	// The input of the block

	IOBlock(int n_reference, int n_error){
		reference = new Tensor(n_reference);
		error = new Tensor(n_error);
		anti_noise = NULL;
		// the input must set only after the rest of the system is formed.
	};

	~IOBlock() {
		delete reference;
		delete error;
	}

	virtual void calc_output();
	virtual void clock_tick() {};
};

class ANC: public SubSystem  {
public:
	ANC();
	void Start();
	float real_time;

	int currente_time;
	int number_reference_sensors;
	int number_anti_noise_sources;
	int number_error_sensors;

	IOBlock *io_block;

	SubSampledSystem *anc_dsp;

	LatticeBlk *controller;
	WhiteNoiseGenerator *aditive_noise;
	Sum  *additive_noise_sum;
	Gain *gain1;

	LSLBlk *inverse_cancelation_path_estimate0;
	LatticeBlk *inverse_cancelation_path_estimate;
	Difference *dif1;

	MFIR *anti_noise_filter;
	Delay *anti_noise_filter_delay;
	Delay *inversion_delay;
	Delay *inversion_delay2;

	LSLBlk *lsl1;

	SubSampledSystem *off_line_modelling_dsp;
};

#endif
