#ifndef TENSOR_H
#define TENSOR_H

#include "global.h"

// Implementacao de uma classe de Tensores
//
#include <assert.h>
#include <string.h>

#ifndef ASSERT
#define ASSERT(exp) assert(exp)
#endif

typedef double val_type;

// *****************************************
// Matrix Multiplication
//
// Os elementos estao agrupados em colunas

inline void Element_Multiply(val_type *x1, val_type *x2, val_type *y, int size) {
	for(int i=0; i<size; i++) { 	y[i]=x1[i]*x2[i];	}
}

inline void Matrix_Multiply(val_type *x1, val_type *x2, val_type *y, int x1l, int x1c, int x2c) {
	int i, j, k;
	int &x2l = x1c;

	for(i=0; i<x1l; i++) {
		for(j=0; j<x2c; j++) {
			y[i+j*x1l] = 0;
			for(k=0; k<x1c; k++) {
				y[i+j*x1l] += x1[i+k*x1l]*x2[k+j*x2l];
			}
		}
	}
}

inline void Matrix_MultiplyT(val_type *x1, val_type *x2, val_type *y, int x1l, int x1c, int x2c) {
	// Y = X1^T * X1
	int i, j, k;
	int &x2l = x1l;

	for(i=0; i<x1c; i++) {
		for(j=0; j<x2c; j++) {
			y[i+j*x1c] = 0;
			for(k=0; k<x1l; k++) {
				y[i+j*x1c] += x1[k+i*x1l]*x2[k+j*x2l];
			}
		}
	}
}

class MVectorLink {
public:
	val_type *values;
	int size;
	int channels;
	double &pos(int i, int j) 	{ASSERT(i<size && j<channels); return values[i+j*size];}
	MVectorLink() {values = NULL; size = 0; channels = 0;};
};

class MVector:public MVectorLink {
public:
	~MVector() { delete [] values; values = NULL; }
};

// *****************************************
// Tensor Class

class Tensor {
public:
	Tensor(int s1) {
		dim=1;
		size_vector=new int[1]; size_vector[0]=s1;
		size = getSize();
		values = new val_type[size];
		memset(values,0, size*sizeof(val_type));
		link = false;
	};
	Tensor(int s1, int s2) {
		dim=2;
		size_vector=new int[2]; size_vector[0]=s1; size_vector[1]=s2;
		size = getSize();
		values = new val_type[size];
		memset(values,0, size*sizeof(val_type));
		link = false;
	};
	Tensor(int s1, int s2, int s3) {
		dim=3;
		size_vector=new int[3]; size_vector[0]=s1; size_vector[1]=s2; size_vector[2]=s3;
		size = getSize();
		values = new val_type[size];
		memset(values,0, size*sizeof(val_type));
		link = false;
	};
	Tensor(int s1, int s2, int s3, int s4) {
		dim=4;
		size_vector=new int[4]; size_vector[0]=s1; size_vector[1]=s2; size_vector[2]=s3; size_vector[3]=s4;
		size = getSize();
		values = new val_type[size];
		memset(values,0, size*sizeof(val_type));
		link = false;
	};

	Tensor(int s[], int d) {
		size_vector= new int[d];
		for(int i=0; i<d; i++) size_vector[i]=s[i];
		dim=d;
		size = getSize();
		values = new val_type[size];
		memset(values,0, size*sizeof(val_type));
		link = false;
	}

	Tensor(int s[], int d, val_type *val) {
		size_vector= new int[d];
		for(int i=0; i<d; i++) size_vector[i]=s[i];
		dim=d;
		size = getSize();
		values = val;
		link=true;
	}

	~Tensor() {
		delete [] size_vector;
		if (!link) delete [] values;
	}


	bool link;
	int dim; 
	int *size_vector;
	int size;

	val_type *values;

	int getSize(void) { 
		int i; int size=1;
		for(i=0; i<dim; i++) {
			size *= size_vector[i];
		};
		return size;
	};

	void resize(int s1, int s2, int s3) {
		ASSERT(s1*s2*s3==size);
		delete size_vector;
		dim = 3;
		size_vector = new int[3];
		size_vector[0]=s1; size_vector[1]=s2; size_vector[2]=s3;
	}

	int getSize(int i) { ASSERT(i<dim); return size_vector[i]; }
	int getDim() { return dim; }

	val_type &pos(int *ind) {
		// A rotina nao e' muito eficiente
		// Nao deve ser utilizada com frequecia
		int n; int i=0;
		int peso = 1;
		
		for(n=0; n<getDim(); n++) { 
			ASSERT(ind[n]<size_vector[n]);
			i += peso*ind[n]; 
			peso *= size_vector[n];
		}
		return values[i];
	};

	// O indice de menor peso e' o da esquerda (ordenado por linhas)!

	val_type &pos(int i) {
		return values[i];
	};

	val_type &pos(int i, int j) {
		return values[j*getSize(0)+i];
	};

	val_type &pos(int i, int j, int k) {
		return values[k*getSize(1)*getSize(0)+j*getSize(0)+i];
	};

	Tensor *Transpose(){
		int *ind = new int[dim];
		int i,n=0;
		Tensor *result;
		bool next_indice;

		// Create new Tensor
		for(i=0; i<dim; i++) ind[i]=size_vector[dim-i-1];
		result = new Tensor(ind, dim);

		// Create new Tensor
		for(i=0; i<dim; i++) ind[i]=0;

		while(n<size){
			result->values[n]=pos(ind);
			n++;
			if (n<size){
				// calculate next index
				i=dim-1;
				do {
					next_indice=false;
					ASSERT(i>=0);
					(ind[i])++;
					if (ind[i]>=size_vector[i]) {
						ind[i]=0;
						i--;
						next_indice = true;
					}
				} while (next_indice);
			}
		}
		delete ind;
		return result;
	}
}; // class

inline void Dot(Tensor &x1, 
		 Tensor &x2, int n, 
		 Tensor &y) {

	int x1l=1;
	int x1c=1;
	int x2c=1;
	int i;

	ASSERT(x1.getDim()+x2.getDim()-2*n == y.getDim());

	for(i=0; i<x1.getDim()-n; i++) {
		ASSERT(x1.getSize(i)==y.getSize(i));
		x1l*=x1.getSize(i);
	};
	for(i=x1.getDim()-n; i<x1.getDim(); i++) {
		x1c*=x1.getSize(i);
		ASSERT(x1.getSize(i)==x2.getSize(x1.getDim()-i-1));
	};
	for(i=n; i<x2.getDim(); i++) {
		ASSERT(x2.getSize(i)==y.getSize(i+x1.getDim()-2*n));
		x2c*=x2.getSize(i);
	};

	Matrix_Multiply(x1.values, x2.values, y.values, x1l, x1c, x2c);
}

inline void Dot(Tensor *x1, Tensor *x2, int n, Tensor *y) {
	Dot(*x1, *x2, n, *y);
}

inline void DotT(Tensor &x1, 
		 Tensor &x2, int n, 
		 Tensor &y) {

	int x1l=1;
	int x1c=1;
	int x2c=1;
	int i;

	ASSERT(x1.getDim()+x2.getDim()-2*n == y.getDim());

	for(i=0; i<n; i++) {
		ASSERT(x1.getSize(i)==x2.getSize(i));
		x1l*=x1.getSize(i);
	};
	for(i=n; i<x1.getDim(); i++) {
		x1c*=x1.getSize(i);
		ASSERT(x1.getSize(i)==y.getSize(x1.getDim()-i-1));
	};
	for(i=n; i<x2.getDim(); i++) {
		ASSERT(x2.getSize(i)==y.getSize(i+x1.getDim()-2*n));
		x2c*=x2.getSize(i);
	};

	Matrix_MultiplyT(x1.values, x2.values, y.values, x1l, x1c, x2c);
}

inline void DotT(Tensor *x1, Tensor *x2, int n, Tensor *y) {
	DotT(*x1, *x2, n, *y);
}

inline void Plus(Tensor &x1, 
		 Tensor &x2, 
		 Tensor &y) {
	int i; int size = x1.getSize();

	#if _DEBUG
	ASSERT(x1.getDim()==x2.getDim() && x2.getDim()==y.getDim());
	for(int d=0; d<x1.getDim(); d++) {
		ASSERT(x1.getSize(d)==x2.getSize(d) && x2.getSize(d)==y.getSize(d));
	}
	#endif

	for(i=0; i<size; i++) {
		y.values[i]=x1.values[i]+x2.values[i];
	}
}

inline void Plus(Tensor *x1, Tensor *x2, Tensor *y) {
	Plus(*x1, *x2, *y);
}

inline void Minus(Tensor &x1, 
		 Tensor &x2, 
		 Tensor &y) {
	int i; int size = x1.getSize();

	#if _DEBUG
	ASSERT(x1.getDim()==x2.getDim() && x2.getDim()==y.getDim());
	for(int d=0; d<x1.getDim(); d++) {
		ASSERT(x1.getSize(d)==x2.getSize(d) && x2.getSize(d)==y.getSize(d));
	}
	#endif

	for(i=0; i<size; i++) {
		y.values[i]=x1.values[i]-x2.values[i];
	}
}

inline void Minus(Tensor *x1, Tensor *x2, Tensor *y) {
	Minus(*x1, *x2, *y);
}

inline void Times(val_type k, 
		 Tensor &x,
		 Tensor &y) {
	int i; int size = x.getSize();

	#if _DEBUG
	ASSERT(x.getDim()==y.getDim());
	for(int d=0; d<x.getDim(); d++) {
		ASSERT(x.getSize(d)==y.getSize(d));
	}
	#endif

	for(i=0; i<size; i++) {
		y.values[i]=k*x.values[i];
	}
}

inline void Times(val_type k, Tensor *x, Tensor *y) {
	Times(k, *x, *y);
}

inline val_type calc_power(Tensor &x) {
	val_type res=0;
	for(int i=0; i<x.getSize(); i++)
		res+= x.values[i]*x.values[i];
	return res;
}

inline val_type calc_power(Tensor *x) {
	return calc_power(*x);
}

#endif // TENSOR_H
